# rwiki.rb
#
# Copyright (c) 2000-2002 Masatoshi SEKI
#
# rwiki.rb is copyrighted free software by Masatoshi SEKI.
# You can redistribute it and/or modify it under the same term as Ruby.

require 'monitor'
require 'rd/rdfmt'
require 'rwiki/rd2rwiki-lib'
require 'rwiki/rw-lib'
require 'rwiki/hotpage'
require 'drb/drb'
require 'erb'
require 'md5'

module RWiki

  module TaintMonitor
    def taint_monitor
      self.taint
      @mon_entering_queue.taint
      @mon_waiting_queue.taint
    end
  end

  Version.regist('rwiki server', '2.0 ext')

  class PageStore
    include Enumerable

    def []=(*arg)
      k = arg.shift
      v = arg.pop
      rev = arg.shift
      opt = arg.shift
      check_revision(k, rev)
      set(k, store(v))
    end

    def [](k)
      retrieve(get(k))
    end

    def modifed(k)
      nil
    end

    def revision(k)
      modified(k).to_s
    end

    def each
      ;
    end

    def check_revision(k, rev)
      return unless rev
      raise RevisionError, [rev, revision(k)].inspect unless rev == revision(k)
    end

    private
    def store(s)
      s
    end

    def retrieve(s)
      s
    end
  end

  class PageStoreFile < PageStore
    include MonitorMixin

    def initialize(path)
      @dir = path.split(File::Separator)
      super()
    end

    private
    def set(k, v, opt=nil)
      synchronize do
	return if v.nil?
	File.open(fname(k), 'w') {|fp| fp.write(v) }
      end
    end

    def get(k)
      synchronize do
	begin
	  File.open(fname(k), 'r') {|fp| fp.read}
	rescue Errno::ENOENT
	  File.open(fname_old(k), 'r') {|fp| fp.read} rescue nil
	end
      end
    end

    public
    def modified(k)
      begin
	File.stat(fname(k)).mtime
      rescue Errno::ENOENT
	File.stat(fname_old(k)).mtime rescue nil
      end
    end

    def revision(k)
      make_digest(get(k))
    end

    def each
      Dir[File.join(@dir + ['*.rd'])].collect do |filename|
	yield(unescape(File.basename(filename, '.rd')))
      end
    end

    private
    def escape(str)
      str.gsub(/([^a-zA-Z0-9_-])/n){ sprintf("%%%02X", $1.unpack("C")[0]) }
    end

    def unescape(str)
      str.tr('+', ' ').gsub(/((?:%[0-9a-fA-F]{2})+)/n) do
	[$1.delete('%')].pack('H*')
      end
    end

    def fname_old(k)
      File.join(*(@dir + [escape(k)]))
    end

    def fname(k)
      fname_old(k) + '.rd'
    end

    def make_digest(src)
      MD5.new(src).hexdigest
    end
  end

  class Content
    include ERB::Util

    EmptyERB = ERB.new('')

    def initialize(name, src)
      @name = name
      @body = nil
      @links = []
      @src = src
      @body_erb = EmptyERB
      set_src(src)
    end
    attr_reader(:name, :body, :body_erb, :links, :src, :tree)
    
    private
    def set_src(src)
      if src
	begin
	  @src = src
	  @tree = RD::RDTree.new("=begin\n#{@src}\n=end\n")
	  v = RD::RD2RWikiVisitor.new
	  @body = v.visit(@tree)
	  @links = v.links
	rescue
	  @body = "<h1>RD Error</h1>\n"
	  @body << "<pre>#{h($!)}</pre>\n"
	  @body << "<pre>\n"
	  cnt = 2	# '=begin' is the first line.
	  @src.each_line do |line|
	    @body << "%4d| %s" % [ cnt, h(line) ]
	    cnt += 1
	  end
	  @body << "</pre>\n"
	end
      end
      @body_erb = ERB.new(@body.to_s)
    end

  end

  class ERbLoader
    def initialize(method_name, fname, dir=nil)
      @dir = dir
      @fname = fname
      @method_name = method_name
    end

    def load(mod)
      reload(mod)
    end

    def reload(mod)
      fname = build_fname(@fname, @dir)
      erb = File.open(fname) {|f| ERB.new(f.read)}
      erb.def_method(mod, @method_name, fname)
    end
    
    private
    def build_fname(fname, dir)
      case dir
      when String
	ary = [dir]
      when Array
	ary = dir
      else
	ary = $:
      end

      found = fname # default
      ary.each do |dir|
	path = File::join(dir, fname)
	if File::readable?(path)
	  found = path
	  break
	end
	path = File::join(dir, 'rwiki', fname)
	if File::readable?(path)
	  found = path
	  break
	end
      end
      found
    end
  end    

  class PageFormat
    include ERB::Util

    @@address = ADDRESS
    @@mailto = MAILTO
    @@css = CSS
    @@title = TITLE
    @@lang = LANG || KCode.lang
    @@charset = CHARSET || KCode.charset

    def navi_view(title, pg)
      %Q[<a href="#{ ref_name(pg.name) }">#{ h title }</a>]
    end

    def modified(t)
      return '-' unless t
      dif = (Time.now - t).to_i
      dif = dif / 60
      return "#{dif}m" if dif <= 60
      dif = dif / 60
      return "#{dif}h" if dif <= 24
      dif = dif / 24
      return "#{dif}d"
    end

    def initialize(env = {}, &block)
      @env = env
      @block = block
    end

    def var(key)
      @block ? @block.call(key) : nil
    end

    def env(key)
      @env[key]
    end

    def ref_url(url)
      h(url)
    end

    def ref_name(name, cmd = 'view')
      page_url =
	if env('ref_name')
	  env('ref_name').call(cmd, name)
	else
	  program = env('base')
	  req = Request.new(cmd, name)
	  program + "?" + req.query
	end
      ref_url(page_url)
    end

    def full_ref_name(name, cmd = 'view')
      page_url = "#{env('base_url')}?#{Request.new(cmd, name).query}"
      ref_url(page_url)
    end

    def form_action
      env('base')
    end

    def form_hidden(name, cmd = 'view')
      # req = Request.new(cmd, name) ???
      %Q[<input type="hidden" name="cmd" value="#{cmd}" />] +
        %Q[<input type="hidden" name="name" value="#{h(name)}" />]
    end

    def body(pg, opt = {})
      str = pg.body_erb.result(binding)
      if opt.has_key?(:key)
	# Copy keys from UI side.
	keys = opt[:key].collect { |i| i.dup }
	str = hilighten(str, keys)
      end
      str
    end

    @rhtml = {}
    @rhtml[:header] = ERbLoader.new('header(pg)', 'header.rhtml')
    @rhtml[:navi] = ERbLoader.new('navi(pg)', 'navi.rhtml')
    @rhtml[:footer] = ERbLoader.new('footer(pg)', 'footer.rhtml')
    @rhtml[:view] = ERbLoader.new('view(pg)', 'view.rhtml')
    @rhtml[:edit] = ERbLoader.new('edit(pg)', 'edit.rhtml')
    @rhtml[:submit] = ERbLoader.new('submit(pg)', 'submit.rhtml')
    @rhtml[:emphasize] = ERbLoader.new('emphasize(pg)', 'emphasize.rhtml')
    @rhtml[:error] = ERbLoader.new('error(pg)', 'err.rhtml')
    @rhtml[:src] = ERbLoader.new('src(pg)', 'src.rhtml')


    def self.reload_rhtml
      @rhtml.each do |k, v|
	v.reload(self)
      end
    end

    reload_rhtml

    private
    def hilighten(str, keywords)
      hilighted = str.dup
      keywords.each do |key|
	re = Regexp.new('(' << Regexp.escape(h(key)) << ')', Regexp::IGNORECASE)
	hilighted.gsub!(/([^<]*)(<[^>]*>)?/) {
	  body, tag = $1, $2
	  body.gsub(re) {
	    %Q[<em style="color:black;background-color:#ffff66">#{$1}</em>]
	  } << ( tag || "" )
	}
      end
      hilighted
    end

    def erb_result(pg, erb_or_str, opt={})
      if String === erb_or_str
	str = erb_or_str
	return str unless str.include?('<%')
	erb = ERB.new(str)
      else
	erb = erb_or_str
      end
      erb.result(binding)
    end
  end

  class SearchFormat < PageFormat
    @rhtml = { :view => RWiki::ERbLoader.new('view(pg)', 'search.rhtml')}
    reload_rhtml
  end

  class Page
    include DRbUndumped

    def initialize(name, book, section)
      @section = section
      @book = book
      @name = name
      @src = nil
      @body_erb = nil
      @links = []
      @revlinks = []
      @modified = db.modified(name)
      @format = PageFormat

      @hot_order = self.method(:hot_order)
      @hot_links = HotPage.new( &@hot_order )
      @hot_revlinks = HotPage.new( &@hot_order )
      @hot_edges = HotPageContainer.new( &@hot_order )
      @hot_edges << @hot_links << @hot_revlinks
    end
    attr_reader(:name, :src, :links, :revlinks, :modified)
    attr_reader(:body_erb)
    attr_reader(:book, :section)
    attr_accessor(:format)

    def revision
      db.revision(@name).to_s
    end

    def src=(v)
      set_src(v, nil)
    end

    def set_src(v, rev, opt=nil)
      @book.synchronize do
	@book.dirty
	db[@name, rev, opt] = v
	@modified = Time.now
	update_src(v)
      end
      @book.gc
    end

    def empty?
      src.nil? || src == ''
    end

    def contain?(str)
      return false if empty?
      src.downcase.index(str.downcase) ? true : false
    end

    def match?(regexp)
      return false if empty?
      regexp =~ src
    end

    def view_html(env = {}, &block)
      @format.new(env, &block).view(self)
    end

    def edit_html(env = {}, &block)
      @format.new(env, &block).edit(self)
    end

    def submit_html(env = {}, &block)
      @format.new(env, &block).submit(self)
    end

    def emphatic_html(env = {}, &block)
      @format.new(env, &block).emphasize(self)
    end

    def error_html(env = {}, &block)
      @format.new(env, &block).error(self)
    end

    def src_html(env = {}, &block)
      @format.new(env, &block).src(self)
    end
    
    def body_html(env = {}, &block)
      @format.new(env, &block).body(self)
    end

    def hot_links
      @hot_links.pages.uniq
    end

    def hot_revlinks
      @hot_revlinks.pages.uniq
    end

    def hot_edges
      @hot_edges.pages.uniq
    end

    # For backward compatibility.
    def hotlinks
      hot_edges.collect { |name| @book[name] }
    end

    def next_pages
      revlinks.collect { |name|
	pg = @book[name]
	pg.links[pg.links.index(@name) + 1]
      }.compact.uniq
    end

    def prev_pages
      revlinks.collect { |name|
	pg = @book[name]
	index = pg.links.index(@name)
	(index > 0) ? pg.links[index - 1] : nil
      }.compact.uniq
    end

    def make_content(v)
      Content.new(@name, v)
    end

    def update_src(v)
      obsolete_links

      content = make_content(v)
      @src = content.src
      # @body_erb = ERB.new(@body, 4)
      @body_erb = content.body_erb
      @links = content.links
      @prop = @section.load_prop(content)
      @hot_links.replace(@links)

      update_links
    end

    def prop(key)
      return nil unless @prop
      @prop[key]
    end

    protected
    def add_link(from)
      @hot_links.set_dirty
    end

    def del_link(from)
    end

    def add_rev_link(from)
      unless @revlinks.include? from
	@revlinks.push from
	@hot_revlinks << from
      end
    end

    def del_rev_link(from)
      @revlinks.delete(from)
      @hot_revlinks.delete(from)
    end

    def db
      @section.db
    end

    private
    def update_links
      @links.each do |to|
	@book[to].add_rev_link(@name) unless to == @name
      end
      @revlinks.each do |from|
	@book[from].add_link(@name) unless from == @name
      end
    end

    def obsolete_links
      @links.each do |to|
	@book[to].del_rev_link(@name) unless to == @name
      end
      @revlinks.each do |from|
	@book[from].del_link(@name) unless from == @name
      end
    end

    def <=>(rhs)
      @name <=> rhs.name
    end

    Paste = Time.at(0)
    def hot_order( a, b )
      a = @book[a].modified || Paste
      b = @book[b].modified || Paste
      b <=> a
    end
  end

  class Front
    include DRbUndumped

    def initialize(book)
      @book = book
    end

    def include?(name)
      @book.include_name?(name)
    end

    def find(str)
      @book.find_body(str).collect do |pg|
        pg.name
      end
    end

    def find_all(str)
      @book.find_body(str).each do |pg|
        yield pg
      end
    end

    def view(name, env = {}, &block)
      page = @book[name]
      page.empty? ? page.edit_html(env, &block) : page.view_html(env, &block)
    end

    def edit_view(name, env = {}, &block)
      @book[name].edit_html(env, &block)
    end

    def submit_view(name, env = {}, &block)
      @book[name].submit_html(env, &block)
    end

    def emphatic_view(name, env = {}, &block)
      @book[name].emphatic_html(env, &block)
    end

    def error_view(name, env = {}, &block)
      @book[name].error_html(env, &block)
    end

    def src(name)
      page = @book[name]
      page.empty? ? nil : page.src
    end

    def body(name)
      page = @book[name]
      return nil if page.empty?
      c = Content.new(page.name, page.src)
      c.body
    end
    
    def modified(name)
      @book[name].modified
    end

    def src_view(name, env = {}, &block)
      @book[name].src_html(env, &block)
    end
    
    def set_src_and_view(name, src, env = {}, &block)
      page = @book[name]
      page.src = src
      page.view_html(env, &block)
    end

    def page(name)
      @book[name]
    end

    def [](name)
      @book[CGI.unescape(name)]
    end

    def default_url( env = {} )
      RWiki::Request.default_url(env)
    end
    
    def rd2content( src )
      return {} if src.to_s.empty?
      c = Content.new('tmp', src)
      return { :body => c.body, :links => c.links }
    end
  end

  module BookConfigMixin
    def initialize
      @db = nil
      @format = nil
      @page = nil
      @prop_hook = []
      @default_src_proc = []
    end
    
    def init_with_config(config)
      @db = config.db
      @format = config.format
      @page = config.page
      @prop_hook = config.prop_hook.dup
      @default_src_proc = config.default_src_proc.dup
    end
    
    def add_prop_loader(key, loader)
      @prop_hook.push([key, loader])
    end

    def add_default_src_proc(src_proc)
      @default_src_proc.unshift(src_proc)
    end
    attr_accessor :db, :format, :page, :prop_hook, :default_src_proc
  end

  class BookConfig
    include BookConfigMixin

    def initialize(config=@@default)
      super()
      init_with_config(config) if config
    end

    @@default = self.new(nil)

    def self.default; @@default; end
  end

  BookConfig.default.db = PageStoreFile.new(DB_DIR)
  BookConfig.default.format = PageFormat
  BookConfig.default.page = Page
  BookConfig.default.add_default_src_proc(proc {|name| "= #{name}\n\n"})

  class Section
    include BookConfigMixin

    def initialize(config, regex = nil)
      super()
      init_with_config(config || BookConfig.default)
      if String === regex
	regex = Regexp.new("^#{Regexp.escape(regex)}$")
      end
      @pattern = regex
    end

    def match?(name)
      return true if @pattern.nil?
      @pattern =~ name
    end
    
    def create_page(name, book)
      pg = @page.new(name, book, self)
      pg.format = @format
      pg
    end
    
    def default_src(name)
      @default_src_proc.each do |src_proc|
	it = src_proc.call(name)
	return it unless it.nil?
      end
      ""
    end

    def orphan(book)
      @db.find_all { |name|
	if book.include_name?(name) && book[name].revlinks.size > 0
	  false
	else
	  true
	end
      }
    end
    
    def load_prop(content)
      return nil if content.src.nil?
      result = {}
      @prop_hook.each do |key, loader|
	result[key] = loader.load(content)
      end
      result
    end
  end

  class Book
    include BookConfigMixin
    include Enumerable
    include MonitorMixin
    include TaintMonitor

    @@section_list = []
    def self.section_list; @@section_list; end

    def initialize(config=BookConfig.default, section_list=@@section_list)
      super()
      taint_monitor()

      @root_section = Section.new(config)
      @section_list = section_list

      @dirty_count = 0
      @extra_page = {}
      @fw_table = {}
      @navi = []

      @toplevel = true
      @pending = []

      disable_gc

      yield(self) if block_given?
      
      RWiki::PageModule.each do |name, format, title|
	install_page_module(name, format, title)
      end
      
      protect_page('top')
      protect_page('help')
      install_page_module('search', SearchFormat)
      
      enable_gc
    end
    attr_reader :navi

    def section(name)
      @section_list.each do |sec|
	return sec if sec.match?(name)
      end
      return @root_section
    end

    def create_page(name)
      section(name).create_page(name, self)
    end

    def [](name)
      sec = section(name)
      synchronize do 
	obj = @fw_table[name]
	return obj if obj
	obj = create_page(name)
	@fw_table[name] = obj
	toplevel = @toplevel
	@toplevel = false
	if toplevel
	  obj.update_src(sec.db[name])
	  while pg = @pending.pop
	    pg.update_src(sec.db[pg.name]) if pg.empty?
	  end
	else
	  @pending.push(obj)
	end
	@toplevel = toplevel
	obj
      end
    end

    def each
      synchronize do
	@fw_table.each_value do |page|
	  yield page
	end
      end
    end

    def orphan
      ary = @root_section.orphan(self)
      @section_list.each do |sec|
	ary.concat(sec.orphan(self))
      end
      ary.uniq.sort
    end

    def size
      @fw_table.size
    end
    
    def find_title(str)
      self.find_all { |page| page.name.downcase.index(str.downcase) }
    end

    # Bare in mind that the argument regexp must be trusted.
    def search_title(regexp)
      self.find_all { |page| regexp =~ page.name }
    end

    def find_body(str)
      self.find_all { |page| page.contain? str }
    end

    # Bare in mind that the argument regexp must be trusted.
    def search_body(regexp)
      self.find_all { |page| page.match? regexp }
    end

    alias find_str find_body

    def protect_page(name)
      synchronize do
	pg = self[name]
	@extra_page[name] = pg
	return pg
      end
    end
    alias protect_gc protect_page

    def include_name?(name)
      @fw_table.include?(name)
    end

    def dirty
      @dirty_count += 1
    end

    def install_page_module(name, format, title=nil)
      synchronize do
	pg = protect_gc(name)
	pg.format = format if format
	@navi.push([title, pg]) if title
      end
    end

    def front
      Front.new(self)
    end

    def disable_gc
      @gc = false
    end

    def enable_gc
      @gc = true
    end
    
    def gc
      synchronize do
	return unless @gc
	return if @dirty_count > 5
	@dirty_count = 0
	hash = {}
	@extra_page.each do |name, pg|
	  active_names(hash, name)
	end
	@fw_table = hash
      end
    end

    def active_names(hash, root)
      return if hash[root] 
      return unless @fw_table[root]
      hash[root] = @fw_table[root]
      hash[root].links.each do |name|
	active_names(hash, name) unless hash[name]
      end
    end

    def default_src(name)
      section(name).default_src(name)
    end

    def load_prop(content)
      section(content.name).load_prop(content)
    end
  end

  PageModule = []
  def self.install_page_module(name, format, title=nil)
    PageModule.push([name, format, title])
  end

  def self.reload_rhtml
    ObjectSpace.each_object(Class) do |o|
      o.reload_rhtml if o.ancestors.include?(PageFormat)
    end
  end
end
